# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# pylint: disable=redefined-outer-name,protected-access,broad-except
# Unit tests for source.py

from unittest.mock import MagicMock

import pytest

from apologies.game import GameMode
from apologies.reward import RewardCalculatorV1
from apologies.source import NoOpInputSource, RandomInputSource, RewardV1InputSource, source


class TestFunctions:
    def test_source_unknown(self):
        with pytest.raises(TypeError):
            source("apologies.game.Blah")  # not a valid class

    def test_source_invalid(self):
        with pytest.raises(ValueError):
            source("apologies.engine.Engine")  # valid class, but not a CharacterInputSource

    def test_source_valid(self):
        ris = source("apologies.source.RandomInputSource")
        assert isinstance(ris, RandomInputSource)

        ris = source("RandomInputSource")
        assert isinstance(ris, RandomInputSource)  # if there's no module, we assume "apologies.source"


class TestNoOpInputSource:
    def test_constructor(self):
        nois = NoOpInputSource()  # the contract says there must be a valid zero-args constructor
        assert nois.name == "NoOpInputSource"
        assert nois.fullname == "apologies.source.NoOpInputSource"

    def test_choose_move(self):
        with pytest.raises(NotImplementedError):
            NoOpInputSource().choose_move(GameMode.ADULT, MagicMock(), MagicMock(), MagicMock())


class TestRandomInputSource:
    def test_constructor(self):
        ris = RandomInputSource()  # the contract says there must be a valid zero-args constructor
        assert ris.name == "RandomInputSource"
        assert ris.fullname == "apologies.source.RandomInputSource"

    def test_choose_move(self):
        move1 = MagicMock()
        move2 = MagicMock()
        move3 = MagicMock()
        legal_moves = [move1, move2, move3]
        ris = RandomInputSource()
        for _ in range(100):
            assert ris.choose_move(GameMode.ADULT, MagicMock(), legal_moves, MagicMock()) in legal_moves


class TestRewardV1InputSource:
    def test_constructor(self):
        ris = RewardV1InputSource()  # the contract says there must be a valid zero-args constructor
        assert ris.name == "RewardV1InputSource"
        assert ris.fullname == "apologies.source.RewardV1InputSource"
        assert isinstance(ris.calculator, RewardCalculatorV1)

    def test_choose_move(self):
        view = MagicMock()

        move1 = MagicMock()
        move2 = MagicMock()
        move3 = MagicMock()
        legal_moves = [move1, move2, move3]

        ris = RewardV1InputSource()

        ris.calculator.calculate = MagicMock(side_effect=[200, 300, 100])
        evaluator = MagicMock(side_effect=[1, 2, 3])

        assert ris.choose_move(GameMode.ADULT, view, legal_moves, evaluator) is move2



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

"""
Character input sources.  A character could be a person or could be computer-driven.
"""

import random
from abc import ABC, abstractmethod
from pydoc import locate
from typing import Callable, List, Tuple

import attr

from .game import GameMode, PlayerView
from .reward import RewardCalculatorV1
from .rules import Move


class CharacterInputSource(ABC):

    """
    A generic source of input for a character, which could be a person or could be computer-driven.
    Concrete character input sources must have a valid zero-arguments constructor.
    """

    @property
    def fullname(self) -> str:
        """Get the fully-qualified name of the character input source."""
        return ".".join([type(self).__module__, type(self).__name__])

    @property
    def name(self) -> str:
        """Get the fully-qualified name of the character input source."""
        return type(self).__name__

    @abstractmethod
    def choose_move(
        self, mode: GameMode, view: PlayerView, legal_moves: List[Move], evaluator: Callable[[PlayerView, Move], PlayerView]
    ) -> Move:
        """
        Choose the next move for a character.

        There is always at least one legal move: a forfeit.  Nothing else is legal, so the
        character must choose to discard one card.  In standard mode, there is effectively no
        choice (since there is only one card in play), but in adult mode the character can choose
        which to discard.  If a move has an empty list of actions, then this is a forfeit.

        The source `must` return a move from among the passed-in set of legal moves.  If a source
        returns an illegal move, then a legal move will be chosen at random and executed.  This way,
        a misbehaving source (or a source attempting to cheat) does not get an advantage.  The game
        rules require a player to make a legal move if one is available, even if that move is
        disadvantageous.

        Args:
            mode(GameMode): Game mode
            view(PlayerView): Player-specific view of the game
            legal_moves(List[Move]): The set of legal moves
            evaluator(Callable[[PlayerView, Move], PlayerView]): Function to evaluate a move, returning new state

        Returns:
            Move: The character's next move as described above
        """


@attr.s
class NoOpInputSource(CharacterInputSource):

    """
    A no-op input source, which raises an error if ever used.

    The Apologies library is designed with a synchronous callback model in mind.  If your
    application uses a different model, you may use lower-level methods to interact with
    the game engine directly, rather than getting user input from a callback.  In that case,
    you will use this character input source.  If you get an error, you'll know that you've
    done something wrong.
    """

    def choose_move(
        self, _mode: GameMode, _view: PlayerView, _moves: List[Move], _evaluator: Callable[[PlayerView, Move], PlayerView]
    ) -> Move:
        raise NotImplementedError


@attr.s
class RandomInputSource(CharacterInputSource):

    """
    A source of input for a character which chooses randomly from among legal moves.
    """

    def choose_move(
        self, mode: GameMode, view: PlayerView, legal_moves: List[Move], unused: Callable[[PlayerView, Move], PlayerView]
    ) -> Move:
        """Randomly choose the next move for a character."""
        return random.choice(legal_moves)


# noinspection PyMethodMayBeStatic
@attr.s
class RewardInputSource(CharacterInputSource):

    """
    A source of input for a character which chooses its next move based on a reward calculation.
    """

    @abstractmethod
    def calculate(self, view: PlayerView, move: Move, evaluator: Callable[[PlayerView, Move], PlayerView]) -> Tuple[Move, float]:
        """Calculate the reward associated with a move, returning a tuple of (Move, reward)."""

    def choose_move(
        self, mode: GameMode, view: PlayerView, legal_moves: List[Move], evaluator: Callable[[PlayerView, Move], PlayerView]
    ) -> Move:
        """Choose the next move for a player by evaluating and scoring the available moves."""
        evaluated = [self.calculate(view, move, evaluator) for move in legal_moves]  # calculate a reward for each move
        evaluated.sort(reverse=True, key=lambda e: e[1])  # sort the highest-scoring move to the top
        return evaluated[0][0]  # return the highest-scoring move


# noinspection PyMethodMayBeStatic
class RewardV1InputSource(RewardInputSource):

    """
    A source of input for a character which chooses its next move based on the RewardCalculatorV1.
    """

    calculator = RewardCalculatorV1()

    def calculate(self, view: PlayerView, move: Move, evaluator: Callable[[PlayerView, Move], PlayerView]) -> Tuple[Move, float]:
        """Calculate the reward associated with a move, returning a tuple of (Move, reward)."""
        return move, self.calculator.calculate(evaluator(view, move))


# noinspection PyCallingNonCallable
def source(name: str) -> CharacterInputSource:
    """
    Create a character input source by name.

    As a special case, if the name is not fully-qualified, we will assume "apologies.source".

    Args:
        name(str): Fully-qualified name of the source, like "apologies.source.RandomInputSource"

    Returns:
        CharacterInputSource: An instance of the named source

    Raises:
        ValueError: If the named source does not exist or is not a CharacterInputSource
    """
    if not "." in name:
        name = "apologies.source.%s" % name
    cls = locate(name)
    if not issubclass(cls, CharacterInputSource):  # type: ignore
        raise ValueError("%s is not a CharacterInputSource" % name)
    return cls()  # type: ignore



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

# We'd prefer to disable these only on Windows, but that isn't an option
# pylint: disable=bad-option-value,no-member,no-name-in-module

"""
Implements a quick'n'dirty game-playing demo using curses.
"""

import curses
from curses import endwin
from signal import SIGWINCH, signal
from time import sleep

from .engine import Character, Engine
from .game import GameMode
from .render import render_board
from .source import CharacterInputSource

# Minimum terminal size needed to support the demo
_MIN_COLS = 155
_MIN_ROWS = 70


def _render_hand(player):
    """Return a string describing the cards in a player's hand."""
    if not player.hand:
        return "n/a"
    else:
        return "%s" % [card.cardtype.value for card in sorted(player.hand)]


def _draw(stdscr, board, state, history):
    """Draw the static portions of the screen."""
    stdscr.clear()
    stdscr.border()
    stdscr.refresh()

    board.border()
    board.refresh()

    state.border()
    state.refresh()

    history.border()
    history.refresh()


def _refresh(source, engine, game, delay_sec, stdscr, board, state, history):
    """Refresh the dynamic portions of the screen."""
    _refresh_screen(source, engine, delay_sec, game, stdscr)
    _refresh_board(game, board)
    _refresh_state(source, engine, delay_sec, game, state)
    _refresh_history(game, history)


def _refresh_screen(unused_source, unused_engine, unused_game, unused_delay_sec, stdscr):
    stdscr.border()
    stdscr.addstr(1, 4, "APOLOGIES DEMO")
    stdscr.addstr(1, 138, "CTRL-C TO EXIT")
    stdscr.move(_MIN_ROWS - 2, _MIN_COLS - 2)  # bottom-right corner
    stdscr.refresh()


def _refresh_board(game, board):
    """Refresh the game board display section of the screen."""
    board.clear()
    board.border()

    row = 0
    for line in render_board(game).splitlines():
        board.addstr(row, 1, line)
        row += 1

    board.refresh()


def _refresh_state(source, engine, delay_sec, game, state):
    """Refresh the game state section of the screen."""
    state.clear()
    state.border()

    state.addstr(2, 2, "CONFIGURATION")
    state.addstr(4, 3, "Players..: %d" % engine.players)
    state.addstr(5, 3, "Mode.....: %s" % engine.mode.value)
    state.addstr(6, 3, "Source...: %s" % type(source).__name__)
    state.addstr(7, 3, "Delay....: %s seconds" % delay_sec)
    state.addstr(8, 3, "State....: %s" % engine.state)

    row = 11
    for player in game.players.values():
        state.addstr(row + 0, 2, "%s PLAYER" % player.color.name)
        state.addstr(row + 2, 3, "Hand.....: %s" % _render_hand(player))
        state.addstr(row + 3, 3, "Pawns....:")
        state.addstr(row + 4, 6, "%s" % player.pawns[0])
        state.addstr(row + 5, 6, "%s" % player.pawns[1])
        state.addstr(row + 6, 6, "%s" % player.pawns[2])
        state.addstr(row + 7, 6, "%s" % player.pawns[3])
        row += 10

    state.refresh()


def _refresh_history(game, history):
    """Refresh the game history section of the screen."""
    history.clear()
    history.border()

    row = 1
    for entry in game.history[-9:]:
        history.addstr(row, 2, "%s" % entry)
        row += 1

    history.refresh()


def _main(stdscr, source: CharacterInputSource, engine: Engine, delay_sec: float):
    """Main routine for the Python curses application, intended to be wrapped by curses.wrapper()"""

    board = curses.newwin(55, 90, 3, 3)
    state = curses.newwin(55, 59, 3, 94)
    history = curses.newwin(11, 150, 58, 3)

    # See https://stackoverflow.com/a/57205676/2907667
    def resize(unused_signum=None, unused_frame=None):
        endwin()
        _draw(stdscr, board, state, history)

    signal(SIGWINCH, resize)
    resize()

    while True:  # loop until the user CTRL-C's the application
        if not engine.completed:
            game = engine.play_next()
            _refresh(source, engine, game, delay_sec, stdscr, board, state, history)
        sleep(delay_sec)


def _force_resize(cols: int, rows: int) -> None:
    """
    Force an xterm to resize via a control sequence.
    I've tested that this works in the standard Apple terminal and a Debian xterm.
    See: https://apple.stackexchange.com/a/47841/249172
    """
    print("\u001b[8;%d;%dt" % (rows, cols))
    sleep(0.5)  # wait for the window to finish resizing; if we try to render before it's done, the window gets hosed up


def run_demo(players: int, mode: GameMode, source: CharacterInputSource, delay_sec: float) -> None:
    """
    Run the quick'n'dirty demo in a terminal window.

    Args:
        players(int): Number of players in the game
        mode(GameMode): The game mode
        source(CharacterInputSource): The source to use for choosing player moves
        delay_sec(float): The delay between turns when executing the game
    """
    characters = [Character(name="Player %d" % player, source=source) for player in range(players)]
    engine = Engine(mode=mode, characters=characters)
    engine.start_game()
    _force_resize(_MIN_COLS, _MIN_ROWS)
    curses.wrapper(_main, source, engine, delay_sec)



from .engine import Character, Engine
from .game import Card, CardType, Game, GameMode, History, Pawn, Player, PlayerColor, PlayerView, Position
from .rules import Action, ActionType, Move, Rules
from .source import CharacterInputSource, NoOpInputSource, RandomInputSource, RewardV1InputSource

__all__ = [
    "Action",
    "ActionType",
    "Card",
    "CardType",
    "Character",
    "CharacterInputSource",
    "Engine",
    "Game",
    "GameMode",
    "History",
    "Move",
    "NoOpInputSource",
    "Pawn",
    "Player",
    "PlayerColor",
    "PlayerView",
    "Position",
    "RandomInputSource",
    "RewardV1InputSource",
    "Rules",
]



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# pylint: disable=redefined-outer-name

import os

import pytest

from apologies.game import Game, PlayerColor
from apologies.render import render_board

FIXTURE_DIR = os.path.join(os.path.dirname(__file__), "fixtures/test_render")


@pytest.fixture
def data():
    data = {}
    for f in os.listdir(FIXTURE_DIR):
        p = os.path.join(FIXTURE_DIR, f)
        if os.path.isfile(p):
            with open(p, encoding="utf-8") as r:
                data[f] = r.read()
    return data


# Unit tests for render_board()
class TestRenderBoard:
    def test_empty_2_player_empty(self, data):
        game = Game(playercount=2)
        expected = data["empty2"]
        actual = render_board(game)
        assert expected == actual

    def test_empty_3_player_empty(self, data):
        game = Game(playercount=3)
        expected = data["empty3"]
        actual = render_board(game)
        assert expected == actual

    def test_empty_4_player_empty(self, data):
        game = Game(playercount=4)
        expected = data["empty4"]
        actual = render_board(game)
        assert expected == actual

    def test_home(self, data):
        game = _fill_home()
        expected = data["home"]
        actual = render_board(game)
        assert expected == actual

    def test_safe_03(self, data):
        game = _fill_safe(0)
        expected = data["safe_03"]
        actual = render_board(game)
        assert expected == actual

    def test_safe_14(self, data):
        game = _fill_safe(1)
        expected = data["safe_14"]
        actual = render_board(game)
        assert expected == actual

    def test_top(self, data):
        game = _fill_squares(0, 15)
        expected = data["top"]
        actual = render_board(game)
        assert expected == actual

    def test_right(self, data):
        game = _fill_squares(16, 29)
        expected = data["right"]
        actual = render_board(game)
        assert expected == actual

    def test_bottom(self, data):
        game = _fill_squares(30, 45)
        expected = data["bottom"]
        actual = render_board(game)
        assert expected == actual

    def test_left(self, data):
        game = _fill_squares(46, 59)
        expected = data["left"]
        actual = render_board(game)
        assert expected == actual


# Create a game with all players in home
def _fill_home() -> Game:
    game = Game(playercount=4)
    for color in [PlayerColor.BLUE, PlayerColor.RED, PlayerColor.YELLOW, PlayerColor.GREEN]:
        for pawn in range(4):
            game.players[color].pawns[pawn].position.move_to_home()
    return game


# Create a game with all players in the safe zone
def _fill_safe(start: int) -> Game:
    game = Game(playercount=4)
    for color in [PlayerColor.BLUE, PlayerColor.RED, PlayerColor.YELLOW, PlayerColor.GREEN]:
        for pawn in range(4):
            game.players[color].pawns[pawn].position.move_to_safe(pawn + start)
    return game


# Fill a range of squares on the board with pieces from various players
def _fill_squares(start: int, end: int) -> Game:
    game = Game(playercount=4)
    square = 0
    for pawn in range(4):
        for color in [PlayerColor.BLUE, PlayerColor.RED, PlayerColor.YELLOW, PlayerColor.GREEN]:
            if square + start <= end:
                game.players[color].pawns[pawn].position.move_to_square(square + start)
                square += 1
    return game



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

"""
Run a simulation to see how well different character input sources behave.
"""

from __future__ import annotations  # so we can return a type from one of its own methods

import csv
import statistics
from itertools import combinations_with_replacement
from typing import Dict, List, Optional, Sequence

import attr
import pendulum
from pendulum import DateTime  # type: ignore

from .engine import Character, Engine
from .game import MAX_PLAYERS, MIN_PLAYERS, GameMode, Player
from .source import CharacterInputSource

BASE_HEADERS = [
    "Scenario",
    "Mode",
    "Iterations",
    "Players",
    "Player 1",
    "Player 2",
    "Player 3",
    "Player 4",
    "Median Turns",
    "Overall Mean Turns",
    "Overall Median Duration (ms)",
    "Overall Mean Duration (ms)",
]

SOURCE_HEADERS = [
    "Median Turns",
    "Mean Turns",
    "Median Duration (ms)",
    "Mean Duration (ms)",
    "Wins",
    "Win %",
]


def _mean(data: Sequence[float]) -> Optional[float]:
    """Calculate the mean rounded to 2 decimal places or return None if there is not any data."""
    return round(statistics.mean(data), 2) if data else None


def _median(data: Sequence[float]) -> Optional[float]:
    """Calculate the median rounded to 2 decimal places or return None if there is not any data."""
    return round(statistics.median(data), 2) if data else None


@attr.s
class _Result:

    """Result of a single game within a scenario."""

    start = attr.ib(type=DateTime)
    stop = attr.ib(type=DateTime)
    character = attr.ib(type=Character)
    player = attr.ib(type=Player)


@attr.s
class _Statistics:

    """Scenario statistics for a source."""

    source = attr.ib(type=Optional[str])
    median_turns = attr.ib(type=Optional[float])
    mean_turns = attr.ib(type=Optional[float])
    median_duration = attr.ib(type=Optional[float])
    mean_duration = attr.ib(type=Optional[float])
    wins = attr.ib(type=int)
    win_percent = attr.ib(type=float)

    @staticmethod
    def for_results(name: Optional[str], results: List[_Result]) -> _Statistics:
        in_scope = [result for result in results if name is None or result.character.source.name == name]
        turns = [result.player.turns for result in in_scope]
        durations = [result.stop.diff(result.start).microseconds / 1000 for result in in_scope]  # in milliseconds
        median_turns = _median(turns)
        mean_turns = _mean(turns)
        median_duration = _median(durations)
        mean_duration = _mean(durations)
        wins = len(in_scope)
        win_percent = 0.0 if len(results) == 0 else round(100.0 * (wins / len(results)), 1)
        return _Statistics(name, median_turns, mean_turns, median_duration, mean_duration, wins, win_percent)


@attr.s
class _Analysis:
    """Groups together information analyzed for a scenario."""

    scenario = attr.ib(type=str)
    mode = attr.ib(type=str)
    iterations = attr.ib(type=int)
    players = attr.ib(type=int)
    playernames = attr.ib(type=List[str])
    overall_stats = attr.ib(type=_Statistics)
    source_stats = attr.ib(type=Dict[str, _Statistics])


def _analyze_scenario(
    scenario: int,
    mode: GameMode,
    iterations: int,
    players: int,
    sources: Sequence[CharacterInputSource],
    combination: Sequence[CharacterInputSource],
    results: List[_Result],
) -> _Analysis:
    """Analyze a scenario, generating data that can be written to the CSV file."""
    playernames = [source.name for source in combination] + [""] * (MAX_PLAYERS - len(combination))
    overall_stats = _Statistics.for_results(None, results)
    source_stats = {name: _Statistics.for_results(name, results) for name in sorted(list({source.name for source in sources}))}
    return _Analysis("Scenario %d" % scenario, mode.name, iterations, players, playernames, overall_stats, source_stats)


def _write_header(csvwriter, sources: List[CharacterInputSource]) -> None:  # type: ignore
    """Write the header into the CSV file."""
    headers = BASE_HEADERS[:]
    for name in sorted(list({source.name for source in sources})):
        for column in SOURCE_HEADERS:
            headers += ["%s - %s" % (name, column)]
    csvwriter.writerow(headers)


def _write_scenario(csvwriter, analysis: _Analysis) -> None:  # type: ignore
    """Write analysis results for a scenario into the CSV file."""
    row = [analysis.scenario, analysis.mode, analysis.iterations, analysis.players]
    row += analysis.playernames
    row += [
        analysis.overall_stats.median_turns,
        analysis.overall_stats.mean_turns,
        analysis.overall_stats.median_duration,
        analysis.overall_stats.mean_duration,
    ]
    for stats in analysis.source_stats.values():
        row += [stats.median_turns, stats.mean_turns, stats.median_duration, stats.mean_duration, stats.wins, stats.win_percent]
    csvwriter.writerow(row)


def _run_scenario(prefix: str, iterations: int, engine: Engine) -> List[_Result]:
    """Run a particular scenario, playing a game repeatedly for a set number of iterations."""
    results = []
    for i in range(0, iterations):
        print(" " * 100, end="\r", flush=True)
        print("%siteration %d" % (prefix, i), end="\r", flush=True)
        start = pendulum.now()
        engine.reset()
        engine.start_game()
        while not engine.completed:
            engine.play_next()
        stop = pendulum.now()
        character, player = engine.winner()  # type: ignore
        results.append(_Result(start, stop, character, player))
    return results


# pylint: disable=too-many-locals,line-too-long
def run_simulation(iterations: int, output: str, sources: List[CharacterInputSource]) -> None:
    """
    Run a simulation.

    Args:
        iterations(int): The number of iterations (number of times to play the game)
        output(str): Path to the output file to write
        sources(List[CharacterInputSource]): The source to use for each player in the game
    """
    with open(output, "w", newline="") as csvfile:
        csvwriter = csv.writer(csvfile, quoting=csv.QUOTE_ALL)
        _write_header(csvwriter, sources)

        start = pendulum.now()
        print("Starting simulation at %s, using %d iterations per scenario" % (start.to_datetime_string(), iterations))  # type: ignore

        scenario = 0
        results = []
        for mode in GameMode:
            for players in range(MIN_PLAYERS, MAX_PLAYERS + 1):
                case = 0
                for combination in combinations_with_replacement(sources, players):
                    case += 1
                    scenario += 1
                    prefix = "Scenario %d: %s mode with %d players (case %d): " % (scenario, mode.name, players, case)
                    characters = [Character(name=source.name, source=source) for source in combination]
                    engine = Engine(mode=mode, characters=characters)
                    print(" " * 100, end="\r", flush=True)
                    print("%sstarting" % prefix, end="\r", flush=True)
                    results = _run_scenario(prefix, iterations, engine)
                    print("%sanalyzing" % prefix, end="\r", flush=True)
                    analysis = _analyze_scenario(scenario, mode, iterations, players, sources, combination, results)
                    print("%swriting CSV" % prefix, end="\r", flush=True)
                    _write_scenario(csvwriter, analysis)
                    print("%sdone" % prefix, end="\r", flush=True)

        stop = pendulum.now()
        print(" " * 100, end="\r", flush=True)
        print("Simulation completed after %s at %s" % (stop.diff(start).in_words(), stop.to_datetime_string()))  # type: ignore



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

"""
Stubs the demo, which does not run on Windows.
"""

from .game import GameMode
from .source import CharacterInputSource


# pylint: disable=unused-argument
def run_demo(players: int, mode: GameMode, source: CharacterInputSource, delay_sec: float) -> None:
    """
    Stubs the demo, which does not run on Windows.
    """
    print("The demo is curses-based and does not run on Windows.")



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

"""
Reward calculations.

Version 1 of the reward algorithm was developed by hand following my own mental model for the
strength of a position.  I used the spreadsheet found in notes/reward.xlsx to refine my
thoughts as I developed the algorithm.

Basically, a pawn is worth more the closer it is to home.  It's worth incrementally more in the
safe zone (where it can't be hurt).  There's an additional bonus for winning, to incentivize
the engine to pick a move that ends the game as fast as possible.  The score is calculated by
comparing the player's position relative to the positions of all its opponents.  We want the
engine to pick the move that both maximizes the player's position and also minimizes the
positions of its opponents.

In simulation runs generated by simulation.py, a reward-based character source vastly out
performs a source that picks its moves randomly.  The worst-case scenario is a 4-player
STANDARD mode game between a single reward-based source and 3 random sources, where the
reward-based source wins about 70% of the time.  This is probably because in a STANDARD mode
game, the possible moves in each turn are fairly limited, due to each player picking and
playing the top card off the deck.  This evens the playing field, because it's quite likely
that any player will have no good move on their turn.  In a 4-player ADULT mode game, where the
engine has the opportunity to choose between more possible moves for each turn, a reward-based
source wins more than 98% of the time against 3 random sources. 
"""

from abc import ABC, abstractmethod
from typing import Tuple

from apologies.game import Player, PlayerView
from apologies.rules import BoardRules


class RewardCalculator(ABC):

    """Abstract reward calculator interface, to support multiple reward implementations."""

    @abstractmethod
    def calculate(self, view: PlayerView) -> float:
        """Calculate the reward associated with a player view."""

    @abstractmethod
    def range(self, players: int) -> Tuple[float, float]:
        """Return the range of possible rewards for a game."""


class RewardCalculatorV1(RewardCalculator):
    """Version 1 of the reward calculator."""

    def calculate(self, view: PlayerView) -> float:
        """Calculate the reward associated with an observation."""
        return float(RewardCalculatorV1._reward(view))

    def range(self, players: int) -> Tuple[float, float]:
        """Return the range of possible rewards for a game."""
        return 0.0, float((players - 1) * 400)  # reward is up to 400 points per opponent

    @staticmethod
    def _reward(view: PlayerView) -> int:
        # Reward measures this player's overall game position relative to their opponents
        player_score = RewardCalculatorV1._player_score(view.player)
        opponent_scores = [RewardCalculatorV1._player_score(player) for player in view.opponents.values()]
        reward = (len(view.opponents) * player_score) - sum(opponent_scores)
        return 0 if reward < 0 else reward

    @staticmethod
    def _player_score(player: Player) -> int:
        # There are 3 different incentives, designed to encourage the right behavior
        distance_incentive = RewardCalculatorV1._distance_incentive(player)
        safe_incentive = RewardCalculatorV1._safe_incentive(player)
        winner_incentive = RewardCalculatorV1._winner_incentive(player)
        return distance_incentive + safe_incentive + winner_incentive

    @staticmethod
    def _distance_incentive(player: Player) -> int:
        # Incentive of 1 point for each square closer to home for each of the player's 4 pawns
        distance = sum([BoardRules.distance_to_home(pawn) for pawn in player.pawns])
        return 260 - distance  # 260 = 4*65, max distance for 4 pawns

    @staticmethod
    def _safe_incentive(player: Player) -> int:
        # Incentive of 10 points for each pawn in safe or home
        return sum([10 if pawn.position.home or pawn.position.safe is not None else 0 for pawn in player.pawns])

    @staticmethod
    def _winner_incentive(player: Player) -> int:
        # Incentive of 100 points for winning the game
        return 100 if player.all_pawns_in_home() else 0



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

"""
Utility functionality.
"""
from typing import Any, Generic, List, TypeVar

import attr
import cattr
from pendulum.datetime import DateTime
from pendulum.parser import parse


class CattrConverter(cattr.Converter):
    """
    Cattr converter that knows how to correctly serialize/deserialize DateTime to an ISO 8601 timestamp.
    """

    def __init__(self) -> None:
        super().__init__()  # type: ignore
        self.register_unstructure_hook(DateTime, lambda datetime: datetime.isoformat() if datetime else None)  # type: ignore
        self.register_structure_hook(DateTime, lambda string, _: parse(string) if string else None)


T = TypeVar("T")  # pylint: disable=invalid-name
"""Generic type"""


@attr.s
class CircularQueue(Generic[T]):
    """A circular queue that keeps returning the original entries repeatedly, in order."""

    entries = attr.ib(type=List[T])
    first = attr.ib(type=Any)
    _working = attr.ib(type=List[T])

    @first.default
    def _default_first(self) -> Any:
        return None if not self.entries else self.entries[0]

    @_working.default
    def _default_working(self) -> List[T]:
        if self.first not in self.entries:
            raise ValueError("First entry not found")
        temp = self.entries[:]
        popped = temp.pop(0)
        while popped != self.first:
            popped = temp.pop(0)
        temp.insert(0, popped)
        return temp

    @entries.validator
    def _check_entries(self, _attribute: str, value: int) -> None:
        if not value:
            raise ValueError("Entries must not be empty")

    def next(self) -> T:
        """Get the next entry from the queue."""
        if len(self._working) == 0:
            self._working.extend(self.entries)
        return self._working.pop(0)



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# pylint: disable=no-self-use,protected-access,too-many-locals,too-many-statements

from unittest.mock import MagicMock, call, patch

import pytest

from apologies.game import ADULT_HAND, DECK_SIZE, PAWNS, Card, CardType, Game, GameMode, Pawn, PlayerColor, Position
from apologies.rules import Action, ActionType, BoardRules, Move, Rules

_UUID = MagicMock(return_value=MagicMock(hex="uuid"))  # any call to get a random UUID returns a UUID with hex value "uuid"


class TestAction:
    def test_constructor(self):
        pawn = Pawn(PlayerColor.BLUE, 1, "whatever")
        position = Position().move_to_square(32)
        action = Action(ActionType.MOVE_TO_START, pawn, position)
        assert action.actiontype == ActionType.MOVE_TO_START
        assert action.pawn is pawn
        assert action.position is position


class TestMove:
    @patch("apologies.rules.uuid.uuid4", new=_UUID)
    def test_constructor_uuid(self):
        card = Card(3, CardType.CARD_12)
        actions = [Action(ActionType.MOVE_TO_START, pawn=Pawn(PlayerColor.BLUE, 1, "whatever"))]
        move = Move(card, actions)
        assert move.card is card
        assert move.actions == actions
        assert move.id == "uuid"

    def test_constructor_explicit(self):
        card = Card(3, CardType.CARD_12)
        actions = [Action(ActionType.MOVE_TO_START, pawn=Pawn(PlayerColor.BLUE, 1, "whatever"))]
        move = Move(card, actions, id="whatever")
        assert move.card is card
        assert move.actions == actions
        assert move.id == "whatever"


class TestRules:
    def test_constructor(self):
        rules = Rules(GameMode.STANDARD)
        assert rules.mode == GameMode.STANDARD

    def test_draw_again(self):
        rules = Rules(GameMode.STANDARD)
        for cardtype in CardType:
            if cardtype == CardType.CARD_2:
                assert rules.draw_again(Card(0, cardtype)) is True
            else:
                assert rules.draw_again(Card(0, cardtype)) is False

    def test_start_game_started(self):
        game = MagicMock(started=True)
        rules = Rules(GameMode.STANDARD)
        with pytest.raises(ValueError):
            rules.start_game(game)

    def test_start_game_standard(self):
        game = Game(2)
        rules = Rules(GameMode.STANDARD)
        rules.start_game(game)

        assert game.started is True
        assert len(game.deck._draw_pile) == DECK_SIZE

        for color in [PlayerColor.RED, PlayerColor.YELLOW]:
            assert game.players[color].color == color
            assert len(game.players[color].hand) == 0

    def test_start_game_adult(self):
        game = Game(4)
        rules = Rules(GameMode.ADULT)
        rules.start_game(game)

        assert game.started is True
        assert len(game.deck._draw_pile) == DECK_SIZE - (4 * ADULT_HAND)

        assert game.players[PlayerColor.RED].color == PlayerColor.RED
        assert game.players[PlayerColor.RED].pawns[0].position.square == 4
        assert len(game.players[PlayerColor.RED].hand) == ADULT_HAND

        assert game.players[PlayerColor.YELLOW].color == PlayerColor.YELLOW
        assert game.players[PlayerColor.YELLOW].pawns[0].position.square == 34
        assert len(game.players[PlayerColor.YELLOW].hand) == ADULT_HAND

        assert game.players[PlayerColor.GREEN].color == PlayerColor.GREEN
        assert game.players[PlayerColor.GREEN].pawns[0].position.square == 49
        assert len(game.players[PlayerColor.GREEN].hand) == ADULT_HAND

        assert game.players[PlayerColor.BLUE].color == PlayerColor.BLUE
        assert game.players[PlayerColor.BLUE].pawns[0].position.square == 19
        assert len(game.players[PlayerColor.BLUE].hand) == ADULT_HAND

    @patch("apologies.rules.uuid.uuid4", new=_UUID)
    def test_construct_legal_moves_no_moves_with_card(self):
        card = MagicMock()

        hand1 = MagicMock()
        hand2 = MagicMock()
        hand = [hand1, hand2]

        pawn1 = MagicMock()
        pawn2 = MagicMock()
        player_pawns = [pawn1, pawn2]

        all_pawns = [MagicMock(), MagicMock()]

        card_pawn1_moves = []
        card_pawn2_moves = []
        legal_moves = [card_pawn1_moves, card_pawn2_moves]
        expected_moves = [Move(card, [])]  # result is a forfeit for the only card

        view = MagicMock()
        view.player = MagicMock(color=PlayerColor.RED, hand=hand, pawns=player_pawns)
        view.all_pawns = MagicMock(return_value=all_pawns)

        rules = Rules(GameMode.STANDARD)
        rules._board_rules.construct_legal_moves = MagicMock(side_effect=legal_moves)
        assert rules.construct_legal_moves(view, card=card) == expected_moves

        rules._board_rules.construct_legal_moves.assert_has_calls(
            [call(PlayerColor.RED, card, pawn1, all_pawns), call(PlayerColor.RED, card, pawn2, all_pawns)]
        )

    @patch("apologies.rules.uuid.uuid4", new=_UUID)
    def test_construct_legal_moves_no_moves_no_card(self):
        card = None

        hand1 = MagicMock()
        hand2 = MagicMock()
        hand = [hand1, hand2]

        pawn1 = MagicMock()
        pawn2 = MagicMock()
        player_pawns = [pawn1, pawn2]

        all_pawns = [MagicMock(), MagicMock()]

        hand1_pawn1_moves = []
        hand1_pawn2_moves = []
        hand2_pawn1_moves = []
        hand2_pawn2_moves = []
        legal_moves = [hand1_pawn1_moves, hand1_pawn2_moves, hand2_pawn1_moves, hand2_pawn2_moves]
        expected_moves = [Move(hand1, []), Move(hand2, [])]  # result is a forfeit for all cards in the hand

        view = MagicMock()
        view.player = MagicMock(color=PlayerColor.RED, hand=hand, pawns=player_pawns)
        view.all_pawns = MagicMock(return_value=all_pawns)

        rules = Rules(GameMode.STANDARD)
        rules._board_rules.construct_legal_moves = MagicMock(side_effect=legal_moves)
        assert rules.construct_legal_moves(view, card=card) == expected_moves

        rules._board_rules.construct_legal_moves.assert_has_calls(
            [
                call(PlayerColor.RED, hand1, pawn1, all_pawns),
                call(PlayerColor.RED, hand1, pawn2, all_pawns),
                call(PlayerColor.RED, hand2, pawn1, all_pawns),
                call(PlayerColor.RED, hand2, pawn2, all_pawns),
            ]
        )

    @patch("apologies.rules.uuid.uuid4", new=_UUID)
    def test_construct_legal_moves_with_moves_with_card(self):
        card = MagicMock()

        hand1 = MagicMock()
        hand2 = MagicMock()
        hand = [hand1, hand2]

        pawn1 = MagicMock()
        pawn2 = MagicMock()
        player_pawns = [pawn1, pawn2]

        all_pawns = [MagicMock(), MagicMock()]

        card_pawn1_moves = [
            Move(card, [Action(ActionType.MOVE_TO_START, pawn1)]),
            Move(card, [Action(ActionType.MOVE_TO_START, pawn1)]),
        ]
        card_pawn2_moves = [Move(card, [Action(ActionType.MOVE_TO_POSITION, pawn2), Action(ActionType.MOVE_TO_START, pawn2)])]
        legal_moves = [card_pawn1_moves, card_pawn2_moves]
        expected_moves = [
            Move(card, [Action(ActionType.MOVE_TO_START, pawn1)]),
            Move(card, [Action(ActionType.MOVE_TO_POSITION, pawn2), Action(ActionType.MOVE_TO_START, pawn2)]),
        ]  # result is a list of all returned moves, with duplicates are removed

        view = MagicMock()
        view.player = MagicMock(color=PlayerColor.RED, hand=hand, pawns=player_pawns)
        view.all_pawns = MagicMock(return_value=all_pawns)

        rules = Rules(GameMode.STANDARD)
        rules._board_rules.construct_legal_moves = MagicMock(side_effect=legal_moves)
        assert rules.construct_legal_moves(view, card=card) == expected_moves

        rules._board_rules.construct_legal_moves.assert_has_calls(
            [call(PlayerColor.RED, card, pawn1, all_pawns), call(PlayerColor.RED, card, pawn2, all_pawns)]
        )

    @patch("apologies.rules.uuid.uuid4", new=_UUID)
    def test_construct_legal_moves_with_moves_no_card(self):
        card = None

        hand1 = MagicMock()
        hand2 = MagicMock()
        hand = [hand1, hand2]

        pawn1 = MagicMock()
        pawn2 = MagicMock()
        player_pawns = [pawn1, pawn2]

        all_pawns = [MagicMock(), MagicMock()]

        hand1_pawn1_moves = [
            Move(hand1, [Action(ActionType.MOVE_TO_START, pawn1)]),
            Move(hand1, [Action(ActionType.MOVE_TO_START, pawn1)]),
        ]
        hand1_pawn2_moves = [Move(hand1, [Action(ActionType.MOVE_TO_START, pawn2), Action(ActionType.MOVE_TO_POSITION, pawn2)])]
        hand2_pawn1_moves = [Move(hand2, [Action(ActionType.MOVE_TO_POSITION, pawn1, Position())])]
        hand2_pawn2_moves = [Move(hand2, [Action(ActionType.MOVE_TO_POSITION, pawn2, Position())])]
        legal_moves = [hand1_pawn1_moves, hand1_pawn2_moves, hand2_pawn1_moves, hand2_pawn2_moves]
        expected_moves = [
            Move(hand1, [Action(ActionType.MOVE_TO_START, pawn1)]),
            Move(hand1, [Action(ActionType.MOVE_TO_START, pawn2), Action(ActionType.MOVE_TO_POSITION, pawn2)]),
            Move(hand2, [Action(ActionType.MOVE_TO_POSITION, pawn1, Position())]),
            Move(hand2, [Action(ActionType.MOVE_TO_POSITION, pawn2, Position())]),
        ]  # result is a list of all returned moves, with duplicates are removed

        view = MagicMock()
        view.player = MagicMock(color=PlayerColor.RED, hand=hand, pawns=player_pawns)
        view.all_pawns = MagicMock(return_value=all_pawns)

        rules = Rules(GameMode.STANDARD)
        rules._board_rules.construct_legal_moves = MagicMock(side_effect=legal_moves)
        assert rules.construct_legal_moves(view, card=card) == expected_moves

        rules._board_rules.construct_legal_moves.assert_has_calls(
            [
                call(PlayerColor.RED, hand1, pawn1, all_pawns),
                call(PlayerColor.RED, hand1, pawn2, all_pawns),
                call(PlayerColor.RED, hand2, pawn1, all_pawns),
                call(PlayerColor.RED, hand2, pawn2, all_pawns),
            ]
        )

    def test_execute_move(self):
        rules = Rules(GameMode.STANDARD)
        game = Game(4)
        player = game.players[PlayerColor.RED]

        for color in PlayerColor:
            for pawn in range(PAWNS):
                game.players[color].pawns[pawn].position.move_to_start = MagicMock()
                game.players[color].pawns[pawn].position.move_to_position = MagicMock()

        move = Move(
            MagicMock(),
            actions=[
                Action(ActionType.MOVE_TO_POSITION, MagicMock(color=PlayerColor.RED, index=1), Position().move_to_square(10)),
                Action(ActionType.MOVE_TO_POSITION, MagicMock(color=PlayerColor.YELLOW, index=3), Position().move_to_square(11)),
            ],
            side_effects=[
                Action(ActionType.MOVE_TO_START, MagicMock(color=PlayerColor.BLUE, index=2)),
                Action(ActionType.MOVE_TO_POSITION, MagicMock(color=PlayerColor.GREEN, index=0), Position().move_to_square(12)),
            ],
        )

        rules.execute_move(game, player, move)

        game.players[PlayerColor.RED].pawns[1].position.move_to_position.assert_called_once_with(Position().move_to_square(10))
        game.players[PlayerColor.YELLOW].pawns[3].position.move_to_position.assert_called_once_with(Position().move_to_square(11))
        game.players[PlayerColor.BLUE].pawns[2].position.move_to_start.assert_called_once()
        game.players[PlayerColor.GREEN].pawns[0].position.move_to_position.assert_called_once_with(Position().move_to_square(12))

    def test_evaluate_move(self):
        move = Move(
            MagicMock(),
            actions=[
                Action(ActionType.MOVE_TO_POSITION, MagicMock(color=PlayerColor.RED, index=1), Position().move_to_square(10)),
                Action(ActionType.MOVE_TO_POSITION, MagicMock(color=PlayerColor.YELLOW, index=3), Position().move_to_square(11)),
            ],
            side_effects=[
                Action(ActionType.MOVE_TO_START, MagicMock(color=PlayerColor.BLUE, index=2)),
                Action(ActionType.MOVE_TO_POSITION, MagicMock(color=PlayerColor.GREEN, index=0), Position().move_to_square(12)),
            ],
        )

        game = Game(4)
        view = game.create_player_view(PlayerColor.RED)

        expected = view.copy()
        expected.player.pawns[1].position.move_to_square(10)
        expected.opponents[PlayerColor.YELLOW].pawns[3].position.move_to_square(11)
        expected.opponents[PlayerColor.BLUE].pawns[2].position.move_to_start()
        expected.opponents[PlayerColor.GREEN].pawns[0].position.move_to_square(12)

        assert expected == Rules.evaluate_move(view, move)


RED = PlayerColor.RED
YELLOW = PlayerColor.YELLOW
GREEN = PlayerColor.GREEN
BLUE = PlayerColor.BLUE


def _pawn(color, start=False, home=False, safe=None, square=None):
    position = Position(start=start, home=home, safe=safe, square=square)
    return Pawn(color, 0, "name", position)


class TestPosition:
    def test_constructor(self):
        BoardRules()  # just make sure it doesn't blow up

    def test_distance_to_home(self):
        # distance from home is always 0
        for color in [RED, YELLOW, GREEN, BLUE]:
            assert BoardRules.distance_to_home(_pawn(color, home=True)) == 0

        # distance from start is always 65
        for color in [RED, YELLOW, GREEN, BLUE]:
            assert BoardRules.distance_to_home(_pawn(color, start=True)) == 65

        # distance from within safe is always <= 5
        assert BoardRules.distance_to_home(_pawn(RED, safe=0)) == 5
        assert BoardRules.distance_to_home(_pawn(BLUE, safe=1)) == 4
        assert BoardRules.distance_to_home(_pawn(YELLOW, safe=2)) == 3
        assert BoardRules.distance_to_home(_pawn(GREEN, safe=3)) == 2
        assert BoardRules.distance_to_home(_pawn(GREEN, safe=4)) == 1

        # distance from circle is always 64
        assert BoardRules.distance_to_home(_pawn(RED, square=4)) == 64
        assert BoardRules.distance_to_home(_pawn(BLUE, square=19)) == 64
        assert BoardRules.distance_to_home(_pawn(YELLOW, square=34)) == 64
        assert BoardRules.distance_to_home(_pawn(GREEN, square=49)) == 64

        # distance from square between turn and circle is always 65
        assert BoardRules.distance_to_home(_pawn(RED, square=3)) == 65
        assert BoardRules.distance_to_home(_pawn(BLUE, square=18)) == 65
        assert BoardRules.distance_to_home(_pawn(YELLOW, square=33)) == 65
        assert BoardRules.distance_to_home(_pawn(GREEN, square=48)) == 65

        # distance from turn is always 6
        assert BoardRules.distance_to_home(_pawn(RED, square=2)) == 6
        assert BoardRules.distance_to_home(_pawn(BLUE, square=17)) == 6
        assert BoardRules.distance_to_home(_pawn(YELLOW, square=32)) == 6
        assert BoardRules.distance_to_home(_pawn(GREEN, square=47)) == 6

        # check some arbitrary squares
        assert BoardRules.distance_to_home(_pawn(RED, square=1)) == 7
        assert BoardRules.distance_to_home(_pawn(RED, square=0)) == 8
        assert BoardRules.distance_to_home(_pawn(RED, square=59)) == 9
        assert BoardRules.distance_to_home(_pawn(RED, square=9)) == 59
        assert BoardRules.distance_to_home(_pawn(BLUE, square=0)) == 23
        assert BoardRules.distance_to_home(_pawn(GREEN, square=40)) == 13

    def test_calculate_position_home(self):
        for color in PlayerColor:
            with pytest.raises(ValueError):
                BoardRules()._position(color, Position().move_to_home(), 1)

    def test_calculate_position_start(self):
        for color in PlayerColor:
            with pytest.raises(ValueError):
                BoardRules()._position(color, Position().move_to_home(), 1)

    def test_calculate_position_from_safe(self):
        for color in PlayerColor:
            assert BoardRules()._position(color, Position().move_to_safe(0), 0) == Position().move_to_safe(0)
            assert BoardRules()._position(color, Position().move_to_safe(3), 0) == Position().move_to_safe(3)

        for color in PlayerColor:
            assert BoardRules()._position(color, Position().move_to_safe(0), 1) == Position().move_to_safe(1)
            assert BoardRules()._position(color, Position().move_to_safe(2), 2) == Position().move_to_safe(4)
            assert BoardRules()._position(color, Position().move_to_safe(4), 1) == Position().move_to_home()

        for color in PlayerColor:
            with pytest.raises(ValueError):
                BoardRules()._position(color, Position().move_to_safe(3), 3)
                BoardRules()._position(color, Position().move_to_safe(4), 2)

        for color in PlayerColor:
            assert BoardRules()._position(color, Position().move_to_safe(4), -2) == Position().move_to_safe(2)
            assert BoardRules()._position(color, Position().move_to_safe(1), -1) == Position().move_to_safe(0)

        assert BoardRules()._position(PlayerColor.RED, Position().move_to_safe(0), -1) == Position().move_to_square(2)
        assert BoardRules()._position(PlayerColor.RED, Position().move_to_safe(0), -2) == Position().move_to_square(1)
        assert BoardRules()._position(PlayerColor.RED, Position().move_to_safe(0), -3) == Position().move_to_square(0)
        assert BoardRules()._position(PlayerColor.RED, Position().move_to_safe(0), -4) == Position().move_to_square(59)
        assert BoardRules()._position(PlayerColor.RED, Position().move_to_safe(0), -5) == Position().move_to_square(58)

        assert BoardRules()._position(PlayerColor.BLUE, Position().move_to_safe(0), -1) == Position().move_to_square(17)
        assert BoardRules()._position(PlayerColor.BLUE, Position().move_to_safe(0), -2) == Position().move_to_square(16)

        assert BoardRules()._position(PlayerColor.YELLOW, Position().move_to_safe(0), -1) == Position().move_to_square(32)
        assert BoardRules()._position(PlayerColor.YELLOW, Position().move_to_safe(0), -2) == Position().move_to_square(31)

        assert BoardRules()._position(PlayerColor.GREEN, Position().move_to_safe(0), -1) == Position().move_to_square(47)
        assert BoardRules()._position(PlayerColor.GREEN, Position().move_to_safe(0), -2) == Position().move_to_square(46)

    def test_calculate_position_from_square(self):
        assert BoardRules()._position(PlayerColor.RED, Position().move_to_square(58), 1) == Position().move_to_square(59)
        assert BoardRules()._position(PlayerColor.RED, Position().move_to_square(59), 1) == Position().move_to_square(0)
        assert BoardRules()._position(PlayerColor.RED, Position().move_to_square(54), 5) == Position().move_to_square(59)
        assert BoardRules()._position(PlayerColor.RED, Position().move_to_square(54), 6) == Position().move_to_square(0)
        assert BoardRules()._position(PlayerColor.RED, Position().move_to_square(54), 7) == Position().move_to_square(1)

        for color in PlayerColor:
            assert BoardRules()._position(color, Position().move_to_square(54), 5) == Position().move_to_square(59)
            assert BoardRules()._position(color, Position().move_to_square(54), 6) == Position().move_to_square(0)
            assert BoardRules()._position(color, Position().move_to_square(54), 7) == Position().move_to_square(1)
            assert BoardRules()._position(color, Position().move_to_square(58), 1) == Position().move_to_square(59)
            assert BoardRules()._position(color, Position().move_to_square(59), 1) == Position().move_to_square(0)
            assert BoardRules()._position(color, Position().move_to_square(0), 1) == Position().move_to_square(1)
            assert BoardRules()._position(color, Position().move_to_square(1), 1) == Position().move_to_square(2)
            assert BoardRules()._position(color, Position().move_to_square(10), 5) == Position().move_to_square(15)

        for color in PlayerColor:
            assert BoardRules()._position(color, Position().move_to_square(59), -5) == Position().move_to_square(54)
            assert BoardRules()._position(color, Position().move_to_square(0), -6) == Position().move_to_square(54)
            assert BoardRules()._position(color, Position().move_to_square(1), -7) == Position().move_to_square(54)
            assert BoardRules()._position(color, Position().move_to_square(59), -1) == Position().move_to_square(58)
            assert BoardRules()._position(color, Position().move_to_square(0), -1) == Position().move_to_square(59)
            assert BoardRules()._position(color, Position().move_to_square(1), -1) == Position().move_to_square(0)
            assert BoardRules()._position(color, Position().move_to_square(2), -1) == Position().move_to_square(1)
            assert BoardRules()._position(color, Position().move_to_square(15), -5) == Position().move_to_square(10)

        assert BoardRules()._position(PlayerColor.RED, Position().move_to_square(0), 3) == Position().move_to_safe(0)
        assert BoardRules()._position(PlayerColor.RED, Position().move_to_square(1), 2) == Position().move_to_safe(0)
        assert BoardRules()._position(PlayerColor.RED, Position().move_to_square(2), 1) == Position().move_to_safe(0)
        assert BoardRules()._position(PlayerColor.RED, Position().move_to_square(1), 3) == Position().move_to_safe(1)
        assert BoardRules()._position(PlayerColor.RED, Position().move_to_square(2), 2) == Position().move_to_safe(1)
        assert BoardRules()._position(PlayerColor.RED, Position().move_to_square(2), 6) == Position().move_to_home()
        assert BoardRules()._position(PlayerColor.RED, Position().move_to_square(51), 12) == Position().move_to_safe(0)
        assert BoardRules()._position(PlayerColor.RED, Position().move_to_square(52), 12) == Position().move_to_safe(1)
        assert BoardRules()._position(PlayerColor.RED, Position().move_to_square(58), 5) == Position().move_to_safe(0)
        assert BoardRules()._position(PlayerColor.RED, Position().move_to_square(59), 4) == Position().move_to_safe(0)

        with pytest.raises(ValueError):
            assert BoardRules()._position(PlayerColor.RED, Position().move_to_square(2), 7) == Position().move_to_home()

        assert BoardRules()._position(PlayerColor.BLUE, Position().move_to_square(16), 2) == Position().move_to_safe(0)
        assert BoardRules()._position(PlayerColor.BLUE, Position().move_to_square(17), 1) == Position().move_to_safe(0)
        assert BoardRules()._position(PlayerColor.BLUE, Position().move_to_square(16), 3) == Position().move_to_safe(1)
        assert BoardRules()._position(PlayerColor.BLUE, Position().move_to_square(17), 2) == Position().move_to_safe(1)
        assert BoardRules()._position(PlayerColor.BLUE, Position().move_to_square(17), 6) == Position().move_to_home()
        with pytest.raises(ValueError):
            assert BoardRules()._position(PlayerColor.BLUE, Position().move_to_square(17), 7) == Position().move_to_home()

        assert BoardRules()._position(PlayerColor.YELLOW, Position().move_to_square(31), 2) == Position().move_to_safe(0)
        assert BoardRules()._position(PlayerColor.YELLOW, Position().move_to_square(32), 1) == Position().move_to_safe(0)
        assert BoardRules()._position(PlayerColor.YELLOW, Position().move_to_square(31), 3) == Position().move_to_safe(1)
        assert BoardRules()._position(PlayerColor.YELLOW, Position().move_to_square(32), 2) == Position().move_to_safe(1)
        assert BoardRules()._position(PlayerColor.YELLOW, Position().move_to_square(32), 6) == Position().move_to_home()
        with pytest.raises(ValueError):
            assert BoardRules()._position(PlayerColor.YELLOW, Position().move_to_square(32), 7) == Position().move_to_home()

        assert BoardRules()._position(PlayerColor.GREEN, Position().move_to_square(46), 2) == Position().move_to_safe(0)
        assert BoardRules()._position(PlayerColor.GREEN, Position().move_to_square(47), 1) == Position().move_to_safe(0)
        assert BoardRules()._position(PlayerColor.GREEN, Position().move_to_square(46), 3) == Position().move_to_safe(1)
        assert BoardRules()._position(PlayerColor.GREEN, Position().move_to_square(47), 2) == Position().move_to_safe(1)
        assert BoardRules()._position(PlayerColor.GREEN, Position().move_to_square(47), 6) == Position().move_to_home()
        with pytest.raises(ValueError):
            assert BoardRules()._position(PlayerColor.GREEN, Position().move_to_square(47), 7) == Position().move_to_home()


def _setup_game():
    game = Game(4)
    for color in PlayerColor:
        for pawn in range(PAWNS):
            game.players[color].pawns[pawn].position.move_to_home()
    return game


def _square(pawn, square):
    return Action(ActionType.MOVE_TO_POSITION, pawn, Position().move_to_square(square))


def _safe(pawn, square):
    return Action(ActionType.MOVE_TO_POSITION, pawn, Position().move_to_safe(square))


def _start(pawn):
    return Action(ActionType.MOVE_TO_START, pawn)


def _bump(view, color, index):
    if view.player.color == color:
        return _start(view.player.pawns[index])
    return _start(view.opponents[color].pawns[index])


def _swap(view, pawn, color, index):
    other = view.opponents[color].pawns[index]
    return [_square(pawn, other.position.square), _square(other, pawn.position.square)]


def _home(pawn):
    return Action(ActionType.MOVE_TO_POSITION, pawn, Position().move_to_home())


def _legal_moves(color, game, index, cardtype):
    card = Card(id="test", cardtype=CardType(cardtype))
    view = game.create_player_view(color)
    pawn = view.player.pawns[index]
    rules = BoardRules()
    moves = rules.construct_legal_moves(view.player.color, card, pawn, view.all_pawns())
    return card, pawn, view, moves


class TestLegalMoves:
    @patch("apologies.rules.uuid.uuid4", new=_UUID)
    def test_construct_legal_moves_card_1(self):
        # No legal moves if no pawn in start, on the board, or in safe
        game = _setup_game()
        card, pawn, view, moves = _legal_moves(RED, game, 0, "1")
        assert moves == []

        # Move pawn from start with no conflicts
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_start()
        card, pawn, view, moves = _legal_moves(RED, game, 0, "1")
        assert moves == [Move(card, actions=[_square(pawn, 4)], side_effects=[])]

        # Move pawn from start with conflict (same color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_start()
        game.players[RED].pawns[1].position.move_to_square(4)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "1")
        assert moves == []  # can't start because we have a pawn there already

        # Move pawn from start with conflict (different color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_start()
        game.players[YELLOW].pawns[0].position.move_to_square(4)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "1")
        assert moves == [Move(card, actions=[_square(pawn, 4)], side_effects=[_bump(view, YELLOW, 0)])]

        # Move pawn on board
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "1")
        assert moves == [Move(card, actions=[_square(pawn, 7)], side_effects=[])]

        # Move pawn on board with conflict (same color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        game.players[RED].pawns[1].position.move_to_square(7)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "1")
        assert moves == []  # can't move because we have a pawn there already

        # Move pawn on board with conflict (different color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        game.players[GREEN].pawns[1].position.move_to_square(7)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "1")
        assert moves == [Move(card, actions=[_square(pawn, 7)], side_effects=[_bump(view, GREEN, 1)])]

    @patch("apologies.rules.uuid.uuid4", new=_UUID)
    def test_construct_legal_moves_card_2(self):
        # No legal moves if no pawn in start, on the board, or in safe
        game = _setup_game()
        card, pawn, view, moves = _legal_moves(RED, game, 0, "1")
        assert moves == []

        # Move pawn from start with no conflicts
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_start()
        card, pawn, view, moves = _legal_moves(RED, game, 0, "2")
        assert moves == [Move(card, actions=[_square(pawn, 4)], side_effects=[])]

        # Move pawn from start with conflict (same color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_start()
        game.players[RED].pawns[1].position.move_to_square(4)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "2")
        assert moves == []  # can't start because we have a pawn there already

        # Move pawn from start with conflict (different color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_start()
        game.players[YELLOW].pawns[0].position.move_to_square(4)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "2")
        assert moves == [Move(card, actions=[_square(pawn, 4)], side_effects=[_bump(view, YELLOW, 0)])]

        # Move pawn on board
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "2")
        assert moves == [Move(card, actions=[_square(pawn, 8)], side_effects=[])]

        # Move pawn on board with conflict (same color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        game.players[RED].pawns[1].position.move_to_square(8)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "2")
        assert moves == []  # can't move because we have a pawn there already

        # Move pawn on board with conflict (different color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        game.players[GREEN].pawns[1].position.move_to_square(8)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "2")
        assert moves == [Move(card, actions=[_square(pawn, 8)], side_effects=[_bump(view, GREEN, 1)])]

    @patch("apologies.rules.uuid.uuid4", new=_UUID)
    def test_construct_legal_moves_card_3(self):
        # No legal moves if no pawn on the board, or in safe
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_home()
        card, pawn, view, moves = _legal_moves(RED, game, 0, "3")
        assert moves == []

        # Move pawn on board
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "3")
        assert moves == [Move(card, actions=[_square(pawn, 9)], side_effects=[])]

        # Move pawn on board with conflict (same color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        game.players[RED].pawns[1].position.move_to_square(9)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "3")
        assert moves == []  # can't move because we have a pawn there already

        # Move pawn on board with conflict (different color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        game.players[GREEN].pawns[1].position.move_to_square(9)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "3")
        assert moves == [Move(card, actions=[_square(pawn, 9)], side_effects=[_bump(view, GREEN, 1)])]

    @patch("apologies.rules.uuid.uuid4", new=_UUID)
    def test_construct_legal_moves_card_4(self):
        # No legal moves if no pawn on the board, or in safe
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_home()
        card, pawn, view, moves = _legal_moves(RED, game, 0, "4")
        assert moves == []

        # Move pawn on board
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "4")
        assert moves == [Move(card, actions=[_square(pawn, 2)], side_effects=[])]

        # Move pawn on board with conflict (same color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        game.players[RED].pawns[1].position.move_to_square(2)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "4")
        assert moves == []  # can't move because we have a pawn there already

        # Move pawn on board with conflict (different color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        game.players[GREEN].pawns[1].position.move_to_square(2)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "4")
        assert moves == [Move(card, actions=[_square(pawn, 2)], side_effects=[_bump(view, GREEN, 1)])]

    @patch("apologies.rules.uuid.uuid4", new=_UUID)
    def test_construct_legal_moves_card_5(self):
        # No legal moves if no pawn on the board, or in safe
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_home()
        card, pawn, view, moves = _legal_moves(RED, game, 0, "5")
        assert moves == []

        # Move pawn on board
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "5")
        assert moves == [Move(card, actions=[_square(pawn, 11)], side_effects=[])]

        # Move pawn on board with conflict (same color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        game.players[RED].pawns[1].position.move_to_square(11)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "5")
        assert moves == []  # can't move because we have a pawn there already

        # Move pawn on board with conflict (different color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        game.players[GREEN].pawns[1].position.move_to_square(11)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "5")
        assert moves == [Move(card, actions=[_square(pawn, 11)], side_effects=[_bump(view, GREEN, 1)])]

    @patch("apologies.rules.uuid.uuid4", new=_UUID)
    def test_construct_legal_moves_card_7(self):
        # No legal moves if no pawn on the board, or in safe
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_home()
        card, pawn, view, moves = _legal_moves(RED, game, 0, "7")
        assert moves == []

        # One move available if there is one pawn on the board
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "7")
        assert moves == [Move(card, actions=[_square(pawn, 13)], side_effects=[])]

        # Multiple moves available if there is more than one pawn on the board
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        game.players[RED].pawns[2].position.move_to_square(55)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "7")
        other = view.player.pawns[2]
        assert moves == [
            Move(card, actions=[_square(pawn, 13)], side_effects=[]),  # move our pawn 7
            Move(card, actions=[_square(pawn, 7), _square(other, 1)], side_effects=[]),  # split (1, 6)
            Move(card, actions=[_square(pawn, 8), _square(other, 0)], side_effects=[]),  # split (2, 5)
            Move(card, actions=[_square(pawn, 9), _square(other, 59)], side_effects=[]),  # split (3, 4)
            Move(card, actions=[_square(pawn, 10), _square(other, 58)], side_effects=[]),  # split (4, 3)
            Move(card, actions=[_square(pawn, 11), _square(other, 57)], side_effects=[]),  # split (5, 2)
            Move(card, actions=[_square(pawn, 12), _square(other, 56)], side_effects=[]),  # split (6, 1)
        ]

        # Either half of a move might bump an opponent back to start
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        game.players[RED].pawns[2].position.move_to_square(55)
        game.players[GREEN].pawns[1].position.move_to_square(10)
        game.players[BLUE].pawns[3].position.move_to_square(56)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "7")
        other = view.player.pawns[2]
        assert moves == [
            Move(card, actions=[_square(pawn, 13)], side_effects=[]),  # move our pawn 7
            Move(card, actions=[_square(pawn, 7), _square(other, 1)], side_effects=[]),  # split (1, 6)
            Move(card, actions=[_square(pawn, 8), _square(other, 0)], side_effects=[]),  # split (2, 5)
            Move(card, actions=[_square(pawn, 9), _square(other, 59)], side_effects=[]),  # split (3, 4)
            Move(card, actions=[_square(pawn, 10), _square(other, 58)], side_effects=[_bump(view, GREEN, 1)]),  # split (4, 3)
            Move(card, actions=[_square(pawn, 11), _square(other, 57)], side_effects=[]),  # split (5, 2)
            Move(card, actions=[_square(pawn, 12), _square(other, 56)], side_effects=[_bump(view, BLUE, 3)]),  # split (6, 1)
        ]

        # If either half of the move has a conflict with another pawn of the same color, the entire move is invalidated
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        game.players[RED].pawns[1].position.move_to_square(9)
        game.players[RED].pawns[2].position.move_to_square(55)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "7")
        other1 = view.player.pawns[1]
        other2 = view.player.pawns[2]
        assert moves == [
            Move(card, actions=[_square(pawn, 13)], side_effects=[]),  # move our pawn 7
            Move(card, actions=[_square(pawn, 7), _square(other1, 15)], side_effects=[]),  # split (1, 6)
            Move(card, actions=[_square(pawn, 8), _square(other1, 14)], side_effects=[]),  # split (2, 5)
            Move(card, actions=[_square(pawn, 9), _square(other1, 13)], side_effects=[]),  # split (3, 4)
            Move(card, actions=[_square(pawn, 10), _square(other1, 12)], side_effects=[]),  # split (4, 3)
            Move(card, actions=[_square(pawn, 11), _square(other1, 11)], side_effects=[]),  # split (5, 2)
            Move(card, actions=[_square(pawn, 12), _square(other1, 10)], side_effects=[]),  # split (6, 1)
            Move(card, actions=[_square(pawn, 7), _square(other2, 1)], side_effects=[]),  # split (1, 6)
            Move(card, actions=[_square(pawn, 8), _square(other2, 0)], side_effects=[]),  # split (2, 5)
            # the move for square 9 is disallowed because pawn[1] already lives there, and isn't part of this action
            Move(card, actions=[_square(pawn, 10), _square(other2, 58)], side_effects=[]),  # split (4, 3)
            Move(card, actions=[_square(pawn, 11), _square(other2, 57)], side_effects=[]),  # split (5, 2)
            Move(card, actions=[_square(pawn, 12), _square(other2, 56)], side_effects=[]),  # split (6, 1)
        ]

    @patch("apologies.rules.uuid.uuid4", new=_UUID)
    def test_construct_legal_moves_card_8(self):
        # No legal moves if no pawn on the board, or in safe
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_home()
        card, pawn, view, moves = _legal_moves(RED, game, 0, "8")
        assert moves == []

        # Move pawn on board
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "8")
        assert moves == [Move(card, actions=[_square(pawn, 14)], side_effects=[])]

        # Move pawn on board with conflict (same color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        game.players[RED].pawns[1].position.move_to_square(14)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "8")
        assert moves == []  # can't move because we have a pawn there already

        # Move pawn on board with conflict (different color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        game.players[GREEN].pawns[1].position.move_to_square(14)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "8")
        assert moves == [Move(card, actions=[_square(pawn, 14)], side_effects=[_bump(view, GREEN, 1)])]

    @patch("apologies.rules.uuid.uuid4", new=_UUID)
    def test_construct_legal_moves_card_10(self):
        # No legal moves if no pawn on the board, or in safe
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_home()
        card, pawn, view, moves = _legal_moves(RED, game, 0, "10")
        assert moves == []

        # Move pawn on board
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(5)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "10")
        assert moves == [
            Move(card, actions=[_square(pawn, 15)], side_effects=[]),
            Move(card, actions=[_square(pawn, 4)], side_effects=[]),
        ]

        # Move pawn on board with conflict (same color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(5)
        game.players[RED].pawns[1].position.move_to_square(15)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "10")
        assert moves == [Move(card, actions=[_square(pawn, 4)], side_effects=[])]  # can't move because we have a pawn there already

        # Move pawn on board with conflict (same color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(5)
        game.players[RED].pawns[1].position.move_to_square(4)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "10")
        assert moves == [
            Move(card, actions=[_square(pawn, 15)], side_effects=[])
        ]  # can't move because we have a pawn there already

        # Move pawn on board with conflict (different color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(5)
        game.players[GREEN].pawns[1].position.move_to_square(15)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "10")
        assert moves == [
            Move(card, actions=[_square(pawn, 15)], side_effects=[_bump(view, GREEN, 1)]),
            Move(card, actions=[_square(pawn, 4)], side_effects=[]),
        ]

        # Move pawn on board with conflict (different color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(5)
        game.players[GREEN].pawns[1].position.move_to_square(4)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "10")
        assert moves == [
            Move(card, actions=[_square(pawn, 15)], side_effects=[]),
            Move(card, actions=[_square(pawn, 4)], side_effects=[_bump(view, GREEN, 1)]),
        ]

    @patch("apologies.rules.uuid.uuid4", new=_UUID)
    def test_construct_legal_moves_card_11(self):
        # No legal moves if no pawn on the board, or in safe
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_home()
        card, pawn, view, moves = _legal_moves(RED, game, 0, "11")
        assert moves == []

        # Move pawn on board
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(15)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "11")
        assert moves == [Move(card, actions=[_square(pawn, 26)], side_effects=[])]

        # Move pawn on board with conflict (same color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(15)
        game.players[RED].pawns[1].position.move_to_square(26)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "11")
        assert moves == []  # can't move because we have a pawn there already

        # Move pawn on board with conflict (different color), which also gets us a swap opportunity
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(15)
        game.players[GREEN].pawns[1].position.move_to_square(26)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "11")
        assert moves == [
            Move(card, actions=_swap(view, pawn, GREEN, 1), side_effects=[]),
            Move(card, actions=[_square(pawn, 26)], side_effects=[_bump(view, GREEN, 1)]),
        ]

        # Swap pawns elsewhere on board
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(15)
        game.players[RED].pawns[1].position.move_to_square(32)  # can't be swapped, same color
        game.players[GREEN].pawns[0].position.move_to_start()  # can't be swapped, in start area
        game.players[YELLOW].pawns[0].position.move_to_safe(0)  # can't be swapped, in safe area
        game.players[YELLOW].pawns[3].position.move_to_square(52)  # can be swapped, on board
        game.players[BLUE].pawns[1].position.move_to_square(19)  # can be swapped, on board\
        card, pawn, view, moves = _legal_moves(RED, game, 0, "11")
        assert moves == [
            Move(card, actions=_swap(view, pawn, YELLOW, 3), side_effects=[]),
            Move(card, actions=_swap(view, pawn, BLUE, 1), side_effects=[]),
            Move(card, actions=[_square(pawn, 26)], side_effects=[]),
        ]

    @patch("apologies.rules.uuid.uuid4", new=_UUID)
    def test_construct_legal_moves_card_12(self):
        # No legal moves if no pawn on the board, or in safe
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_home()
        card, pawn, view, moves = _legal_moves(RED, game, 0, "12")
        assert moves == []

        # Move pawn on board
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "12")
        assert moves == [Move(card, actions=[_square(pawn, 18)], side_effects=[])]

        # Move pawn on board with conflict (same color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        game.players[RED].pawns[1].position.move_to_square(18)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "12")
        assert moves == []  # can't move because we have a pawn there already

        # Move pawn on board with conflict (different color)
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(6)
        game.players[GREEN].pawns[1].position.move_to_square(18)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "12")
        assert moves == [Move(card, actions=[_square(pawn, 18)], side_effects=[_bump(view, GREEN, 1)])]

    @patch("apologies.rules.uuid.uuid4", new=_UUID)
    def test_construct_legal_moves_card_apologies(self):
        # No legal moves if no pawn in start
        game = _setup_game()
        card, pawn, view, moves = _legal_moves(RED, game, 0, "A")
        game.players[YELLOW].pawns[3].position.move_to_square(52)  # can be swapped, on board
        game.players[BLUE].pawns[1].position.move_to_square(19)  # can be swapped, on board
        assert moves == []

        # Swap pawns elsewhere on board
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_start()
        game.players[GREEN].pawns[0].position.move_to_start()  # can't be swapped, in start area
        game.players[YELLOW].pawns[0].position.move_to_safe(0)  # can't be swapped, in safe area
        game.players[YELLOW].pawns[3].position.move_to_square(52)  # can be swapped, on board
        game.players[BLUE].pawns[1].position.move_to_square(19)  # can be swapped, on board
        card, pawn, view, moves = _legal_moves(RED, game, 0, "A")
        assert moves == [
            Move(card, actions=[_square(pawn, 52), _bump(view, YELLOW, 3)], side_effects=[]),
            Move(card, actions=[_square(pawn, 19), _bump(view, BLUE, 1)], side_effects=[]),
        ]

    @patch("apologies.rules.uuid.uuid4", new=_UUID)
    def test_construct_legal_moves_special(self):
        # Move pawn into safe zone
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(2)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "1")
        assert moves == [Move(card, actions=[_safe(pawn, 0)], side_effects=[])]

        # Move pawn to home
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_safe(4)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "1")
        assert moves == [Move(card, actions=[_home(pawn)], side_effects=[])]

        # Move pawn past home
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_safe(4)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "2")
        assert moves == []  # No moves, because it isn't legal

        # Slide of the same color
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(8)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "1")
        assert moves == [Move(card, actions=[_square(pawn, 9)], side_effects=[])]

        # Slide of a different color
        game = _setup_game()
        game.players[RED].pawns[0].position.move_to_square(15)
        game.players[RED].pawns[1].position.move_to_square(17)
        game.players[YELLOW].pawns[2].position.move_to_square(18)
        card, pawn, view, moves = _legal_moves(RED, game, 0, "1")
        assert moves == [Move(card, actions=[_square(pawn, 19)], side_effects=[_bump(view, RED, 1), _bump(view, YELLOW, 2)])]
